package ru.dbotthepony.kstarbound.io

import com.google.common.collect.ImmutableList
import com.google.common.collect.ImmutableMap
import com.google.common.collect.Interner
import com.google.common.collect.Interners
import com.google.gson.JsonSyntaxException
import com.google.gson.TypeAdapter
import com.google.gson.internal.bind.TypeAdapters
import com.google.gson.stream.JsonReader
import com.google.gson.stream.JsonToken
import com.google.gson.stream.JsonWriter
import it.unimi.dsi.fastutil.objects.Object2IntArrayMap
import it.unimi.dsi.fastutil.objects.ObjectArraySet
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.getValue
import ru.dbotthepony.kstarbound.setValue
import java.lang.reflect.Constructor
import kotlin.jvm.internal.DefaultConstructorMarker
import kotlin.reflect.*
import kotlin.reflect.full.isSuperclassOf
import kotlin.reflect.full.isSupertypeOf
import kotlin.reflect.full.memberProperties

@Suppress("unchecked_cast")
private fun <T> resolveBound(bound: Class<T>, stringAdapter: TypeAdapter<String>): TypeAdapter<T> {
	return when (bound) {
		Float::class.java -> TypeAdapters.FLOAT as TypeAdapter<T>
		Double::class.java -> TypeAdapters.DOUBLE as TypeAdapter<T>
		String::class.java -> stringAdapter as TypeAdapter<T>
		Int::class.java -> TypeAdapters.INTEGER as TypeAdapter<T>
		Long::class.java -> TypeAdapters.LONG as TypeAdapter<T>
		Boolean::class.java -> TypeAdapters.BOOLEAN as TypeAdapter<T>
		else -> Starbound.gson.getAdapter(bound)
	}
}

private class ListAdapter<T>(private val bound: Class<T>, private val stringAdapter: TypeAdapter<String>) : TypeAdapter<List<T>?>() {
	private val resolvedBound by lazy { resolveBound(bound, stringAdapter) }

	override fun write(out: JsonWriter, value: List<T>?) {
		out.beginArray()

		if (value != null) {
			val resolvedBound = resolvedBound

			for (v in value) {
				resolvedBound.write(out, v)
			}
		}

		out.endArray()
	}

	override fun read(reader: JsonReader): List<T> {
		reader.beginArray()

		val builder = ImmutableList.builder<T>()
		val resolvedBound = resolvedBound

		while (reader.peek() != JsonToken.END_ARRAY) {
			val readObject = resolvedBound.read(reader) ?: throw JsonSyntaxException("List does not accept nulls")
			builder.add(readObject as T)
		}

		reader.endArray()

		return builder.build()
	}
}

private class MapAdapter<K, V>(private val boundKey: Class<K>, private val boundValue: Class<V>, private val stringAdapter: TypeAdapter<String>) : TypeAdapter<Map<K, V>?>() {
	private val resolvedKey by lazy { resolveBound(boundKey, stringAdapter) }
	private val resolvedValue by lazy { resolveBound(boundValue, stringAdapter) }

	override fun write(out: JsonWriter, value: Map<K, V>?) {
		out.beginArray()

		if (value != null) {
			val resolvedKey = resolvedKey
			val resolvedValue = resolvedValue

			for ((k, v) in value) {
				out.beginArray()
				resolvedKey.write(out, k)
				resolvedValue.write(out, v)
				out.endArray()
			}
		}

		out.endArray()
	}

	override fun read(reader: JsonReader): Map<K, V> {
		reader.beginArray()

		val builder = ImmutableMap.builder<K, V>()

		val resolvedKey = resolvedKey
		val resolvedValue = resolvedValue

		while (reader.peek() != JsonToken.END_ARRAY) {
			reader.beginArray()
			builder.put(resolvedKey.read(reader), resolvedValue.read(reader))
			reader.endArray()
		}

		reader.endArray()

		return builder.build()
	}
}

private class StringMapAdapter<V>(private val bound: Class<V>, private val stringAdapter: TypeAdapter<String>, private val interner: Interner<String>) : TypeAdapter<Map<String, V>?>() {
	private val resolvedBound by lazy { resolveBound(bound, stringAdapter) }

	override fun write(out: JsonWriter, value: Map<String, V>?) {
		val resolvedBound = resolvedBound

		out.beginObject()

		if (value != null) {
			for ((k, v) in value) {
				out.name(k)
				resolvedBound.write(out, v)
			}
		}

		out.endObject()
	}

	override fun read(reader: JsonReader): Map<String, V> {
		val builder = ImmutableMap.builder<String, V>()

		reader.beginObject()

		val resolvedBound = resolvedBound

		while (reader.peek() != JsonToken.END_OBJECT) {
			builder.put(interner.intern(reader.nextName()), resolvedBound.read(reader))
		}

		reader.endObject()

		return builder.build()
	}
}

private class LazyTypeProvider<T : Any?>(private val bound: Class<T>) : TypeAdapter<T>() {
	private val resolved by lazy { resolveBound(bound, TypeAdapters.STRING) }

	override fun write(out: JsonWriter, value: T) {
		resolved.write(out, value)
	}

	override fun read(`in`: JsonReader): T {
		return resolved.read(`in`)
	}
}

private data class PackedProperty<Clazz : Any, T>(
	val property: KProperty1<Clazz, T>,
	val adapter: TypeAdapter<T>,
	val transformer: (T) -> T = { it }
) {
	val returnType = property.returnType
}

/**
 * TypeAdapter для классов которые создаются единожды и более не меняются ("бетонных классов").
 */
class KConcreteTypeAdapter<T : Any> private constructor(
	val bound: KClass<T>,
	private val types: ImmutableList<PackedProperty<T, *>>,
	private val asJsonArray: Boolean = false,
	private val stringInterner: Interner<String>
) : TypeAdapter<T>() {
	private val mapped = Object2IntArrayMap<String>()
	private val loggedMisses = ObjectArraySet<String>()

	var currentSymbolicName by ThreadLocal<String>()

	init {
		mapped.defaultReturnValue(-1)

		for ((i, pair) in types.withIndex()) {
			mapped[pair.property.name] = i
		}
	}

	/**
	 * Обычный конструктор класса (без флагов "значения по умолчанию")
	 */
	private val regularFactory: KFunction<T> = bound.constructors.firstOrNull first@{
		if (it.parameters.size == types.size) {
			val iterator = types.iterator()

			for (param in it.parameters) {
				val nextParam = iterator.next()

				val a = param.type
				val b = nextParam.returnType

				if (!a.isSupertypeOf(b) || a.isMarkedNullable != b.isMarkedNullable) {
					return@first false
				}
			}

			return@first true
		}

		return@first false
	} ?: throw NoSuchElementException("Unable to determine constructor for ${bound.qualifiedName} matching (${types.joinToString(", ")})")

	/**
	 * Синтетический конструктор класса, который создаётся Kotlin'ном, для создания классов со значениями по умолчанию
	 */
	private val syntheticFactory: Constructor<T>? = try {
		bound.java.getDeclaredConstructor(*types.map { (it.returnType.classifier as KClass<*>).java }.also {
			it as MutableList
			it.add(Int::class.java)
			it.add(DefaultConstructorMarker::class.java)
		}.toTypedArray())
	} catch(_: NoSuchMethodException) {
		null
	}

	private val syntheticPrimitives: Array<Any?>?

	init {
		if (syntheticFactory == null) {
			syntheticPrimitives = null
		} else {
			syntheticPrimitives = arrayOfNulls(syntheticFactory.parameters.size)

			for ((i, param) in syntheticFactory.parameters.withIndex()) {
				val type = param.parameterizedType as? Class<*> ?: continue

				if (type.isPrimitive) {
					syntheticPrimitives[i] = when (type) {
						java.lang.Boolean.TYPE -> false
						java.lang.Integer.TYPE -> 0
						java.lang.Long.TYPE -> 0L
						java.lang.Short.TYPE -> (0).toShort()
						java.lang.Byte.TYPE -> (0).toByte()
						java.lang.Character.TYPE -> (0).toChar()
						java.lang.Float.TYPE -> 0f
						java.lang.Double.TYPE -> 0.0
						else -> throw IllegalArgumentException("mamma mia: ${type.name}")
					}
				}
			}
		}
	}

	override fun write(out: JsonWriter, value: T) {
		out.beginObject()

		for ((field, adapter) in types) {
			out.name(field.name)
			@Suppress("unchecked_cast")
			(adapter as TypeAdapter<Any>).write(out, (field as KProperty1<T, Any>).get(value))
		}

		out.endObject()
	}

	override fun read(reader: JsonReader): T {
		if (asJsonArray) {
			reader.beginArray()
		} else {
			reader.beginObject()
		}

		// таблица присутствия значений (если значение true то на i было значение внутри json)
		val presentValues = BooleanArray(types.size)
		val readValues = arrayOfNulls<Any>(types.size)

		// Если нам необходимо читать объект как набор данных массива, то давай
		if (asJsonArray) {
			val iterator = types.iterator()
			var fieldId = 0

			while (reader.peek() != JsonToken.END_ARRAY) {
				if (!iterator.hasNext()) {
					val name = fieldId.toString()

					if (loggedMisses.add(name)) {
						if (currentSymbolicName == null) {
							LOGGER.warn("Skipping JSON field with name $name because ${bound.simpleName} has no such field")
						} else {
							LOGGER.warn("Skipping JSON field with name $name because ${bound.simpleName} has no such field (reading: $currentSymbolicName)")
						}
					}

					reader.skipValue()
					fieldId++

					continue
				}

				val tuple = iterator.next()
				val (field, adapter) = tuple

				try {
					readValues[fieldId] = (tuple.transformer as (Any?) -> Any?)(adapter.read(reader))
					presentValues[fieldId] = true
				} catch(err: Throwable) {
					throw JsonSyntaxException("Exception reading field ${field.name} for ${bound.qualifiedName}", err)
				}

				fieldId++
			}
		// иначе - читаем как json object
		} else {
			while (reader.peek() != JsonToken.END_OBJECT) {
				val name = reader.nextName()
				val fieldId = mapped.getInt(name)

				if (fieldId == -1) {
					if (loggedMisses.add(name)) {
						if (currentSymbolicName == null) {
							LOGGER.warn("Skipping JSON field with name $name because ${bound.simpleName} has no such field")
						} else {
							LOGGER.warn("Skipping JSON field with name $name because ${bound.simpleName} has no such field (reading: $currentSymbolicName)")
						}
					}

					reader.skipValue()
				} else {
					val tuple = types[fieldId]
					val (field, adapter) = tuple

					try {
						readValues[fieldId] = (tuple.transformer as (Any?) -> Any?)(adapter.read(reader))
						presentValues[fieldId] = true
					} catch(err: Throwable) {
						if (currentSymbolicName == null) {
							throw JsonSyntaxException("Exception reading field ${field.name} for ${bound.qualifiedName}", err)
						} else {
							throw JsonSyntaxException("Exception reading field ${field.name} for ${bound.qualifiedName} (reading: $currentSymbolicName)", err)
						}
					}
				}
			}
		}

		if (asJsonArray) {
			reader.endArray()
		} else {
			reader.endObject()
		}

		// если у нас есть все значения для конструктора (все значения присутствуют в исходном json объекте)
		// или у нас нет синтетического конструктора, то делаем вызов "обычного" конструктора
		if (syntheticFactory == null || presentValues.all { it }) {
			for ((i, tuple) in types.withIndex()) {
				val (field) = tuple

				if (readValues[i] == null) {
					if (!tuple.returnType.isMarkedNullable) {
						throw JsonSyntaxException("Field ${field.name} of ${bound.qualifiedName} does not accept nulls")
					}

					if (!regularFactory.parameters[i].isOptional && !presentValues[i]) {
						throw JsonSyntaxException("Field ${field.name} of ${bound.qualifiedName} must be defined (even just as null)")
					}
				}
			}

			@Suppress("unchecked_cast")
			return regularFactory.call(*readValues as Array<out Any>)
		// иначе - в бой вступает синтетический конструктор, с флагами "значения по умолчанию"
		} else {
			// количество bitflag'ов значений по умолчанию
			val ints = if (presentValues.size % 31 == 0) presentValues.size / 31 else presentValues.size / 31 + 1
			// максимум считанных значений + bitflag аргументы значений по умолчанию + null типа DefaultConstructorMarker
			val copied = readValues.copyOf(readValues.size + ints + 1)
			var intIndex = readValues.size
			var target = 0
			var targetMove = 0

			for (bool in presentValues) {
				if (!bool) {
					target = target.or(1.shl(targetMove))
				}

				targetMove++

				if (targetMove >= 32) {
					copied[intIndex++] = target
					target = 0
					targetMove = 0
				}
			}

			if (targetMove != 0) {
				copied[intIndex] = target
			}

			val syntheticPrimitives = syntheticPrimitives!!

			for ((i, tuple) in types.withIndex()) {
				if (copied[i] != null) {
					continue
				}

				val (field) = tuple
				val param = regularFactory.parameters[i]

				if (!param.isOptional && !presentValues[i]) {
					throw JsonSyntaxException("Field ${field.name} of ${bound.qualifiedName} is missing")
				}

				if (tuple.returnType.isMarkedNullable) {
					continue
				}

				if (param.isOptional && !presentValues[i]) {
					copied[i] = syntheticPrimitives[i]
					continue
				}

				throw JsonSyntaxException("Field ${field.name} of ${bound.qualifiedName} does not accept nulls")
			}

			return syntheticFactory.newInstance(*copied)
		}
	}

	/**
	 * Позволяет построить класс [KConcreteTypeAdapter] на основе заданных параметров
	 */
	class Builder<T : Any>(val clazz: KClass<T>) {
		private val types = ArrayList<PackedProperty<T, *>>()
		private val stringInterner = Interners.newWeakInterner<String>()

		private val internedStringAdapter: TypeAdapter<String> = object : TypeAdapter<String>() {
			override fun write(out: JsonWriter, value: String) {
				return TypeAdapters.STRING.write(out, value)
			}

			override fun read(`in`: JsonReader): String {
				return stringInterner.intern(TypeAdapters.STRING.read(`in`))
			}
		}

		/**
		 * Добавляет поле с определённым адаптером
		 */
		fun <V> plain(field: KProperty1<T, V>, adapter: TypeAdapter<V>): Builder<T> {
			types.add(PackedProperty(field, adapter))
			return this
		}

		/**
		 * Добавляет поля без generic типов и без преобразователей
		 */
		@Suppress("unchecked_cast")
		fun plain(vararg fields: KProperty1<T, *>): Builder<T> {
			for (field in fields)
				plain(field)

			return this
		}

		/**
		 * Добавляет поле без generic типов, опционально с преобразователем
		 */
		@Suppress("unchecked_cast")
		fun <In> plain(field: KProperty1<T, In>, transformer: (In) -> In = { it }): Builder<T> {
			val returnType = field.returnType
			val classifier = returnType.classifier as? KClass<*> ?: throw ClassCastException("Unable to cast ${returnType.classifier} to KClass of property ${field.name}!")

			if (classifier.isSuperclassOf(Float::class)) {
				types.add(PackedProperty(field as KProperty1<T, Float>, TypeAdapters.FLOAT, transformer = transformer as (Number) -> Number))
			} else if (classifier.isSuperclassOf(Double::class)) {
				types.add(PackedProperty(field as KProperty1<T, Double>, TypeAdapters.DOUBLE, transformer = transformer as (Number) -> Number))
			} else if (classifier.isSuperclassOf(Int::class)) {
				types.add(PackedProperty(field as KProperty1<T, Int>, TypeAdapters.INTEGER, transformer = transformer as (Number) -> Number))
			} else if (classifier.isSuperclassOf(Long::class)) {
				types.add(PackedProperty(field as KProperty1<T, Long>, TypeAdapters.LONG, transformer = transformer as (Number) -> Number))
			} else if (classifier.isSuperclassOf(String::class)) {
				types.add(PackedProperty(field as KProperty1<T, String>, internedStringAdapter, transformer = transformer as (String) -> String))
			} else if (classifier.isSuperclassOf(Boolean::class)) {
				types.add(PackedProperty(field as KProperty1<T, Boolean>, TypeAdapters.BOOLEAN, transformer = transformer as (Boolean) -> Boolean))
			} else {
				types.add(PackedProperty(field, LazyTypeProvider(classifier.java) as TypeAdapter<Any?>, transformer = transformer as (Any?) -> Any?))
			}

			return this
		}

		/**
		 * Добавляет поле(я) без generic типов.
		 *
		 * Если поле с данным именем не найдено, кидается [NoSuchElementException]
		 */
		fun plain(vararg fields: String): Builder<T> {
			val members = clazz.memberProperties

			return this.plain(*Array(fields.size) {
				val field = fields[it]
				return@Array members.firstOrNull { it.name == field } ?: throw NoSuchElementException("Unable to find field $field")
			})
		}

		/**
		 * Добавляет поле, которое содержит список значений V (без null).
		 *
		 * Список неизменяем (создаётся объект [ImmutableList])
		 */
		fun <V : Any> list(field: KProperty1<T, List<V>?>, type: Class<V>): Builder<T> {
			types.add(PackedProperty(field, ListAdapter(type, internedStringAdapter)))
			return this
		}

		/**
		 * Добавляет поле, которое содержит список значений V (без null).
		 *
		 * Список неизменяем (создаётся объект [ImmutableList])
		 */
		inline fun <reified V : Any> list(field: KProperty1<T, List<V>?>): Builder<T> {
			return this.list(field, V::class.java)
		}

		/**
		 * Добавляет поле-таблицу, которое кодируется как [[key, value], [key, value], ...]
		 *
		 * Таблица неизменяема (создаётся объект [ImmutableMap])
		 */
		fun <K, V> map(field: KProperty1<T, Map<K, V>>, keyType: Class<K>, valueType: Class<V>): Builder<T> {
			types.add(PackedProperty(field, MapAdapter(keyType, valueType, internedStringAdapter)))
			return this
		}

		/**
		 * Добавляет поле-таблицу, которое кодируется как [[key, value], [key, value], ...]
		 *
		 * Таблица неизменяема (создаётся объект [ImmutableMap])
		 */
		fun <K : Any, V : Any> map(field: KProperty1<T, Map<K, V>?>, keyType: KClass<K>, valueType: KClass<V>): Builder<T> {
			types.add(PackedProperty(field, MapAdapter(keyType.java, valueType.java, internedStringAdapter)))
			return this
		}

		/**
		 * Добавляет поле-таблицу, которое кодируется как {"a": value, "b": value, ...}
		 *
		 * Таблица неизменяема (создаётся объект [ImmutableMap])
		 */
		fun <V> map(field: KProperty1<T, Map<String, V>?>, valueType: Class<V>): Builder<T> {
			types.add(PackedProperty(field, StringMapAdapter(valueType, internedStringAdapter, stringInterner)))
			return this
		}

		/**
		 * Добавляет поле-таблицу, которое кодируется как {"a": value, "b": value, ...}
		 *
		 * Таблица неизменяема (создаётся объект [ImmutableMap])
		 */
		fun <V : Any> map(field: KProperty1<T, Map<String, V>?>, valueType: KClass<V>): Builder<T> {
			types.add(PackedProperty(field, StringMapAdapter(valueType.java, internedStringAdapter, stringInterner)))
			return this
		}

		fun build(asList: Boolean = false): KConcreteTypeAdapter<T> {
			return KConcreteTypeAdapter(
				bound = clazz,
				types = ImmutableList.copyOf(types),
				asJsonArray = asList,
				stringInterner = stringInterner
			)
		}
	}

	companion object {
		private val LOGGER = LogManager.getLogger()

		/*fun <T : Any> simple(clazz: KClass<T>, asList: Boolean = false): KConcreteTypeAdapter<T> {
			val builder = Builder(clazz)

			for (argument in clazz.primaryConstructor!!.parameters) {
				builder.plain(argument.)
			}

			return builder.build(asList = asList)
		}*/
	}
}
